#include <lauxlib.h>
#include <lualib.h>

#include <iostream>

#include "window.hpp"

using gol_window = game_window<64, 64, 8>;
using cell_matrix = typename gol_window::cell_matrix;

// Counts how many cells around `cells[x][y]` is alive
//
// WARNING: this function does not check boundaries of coordinates
int liveness(const cell_matrix& cells, std::size_t x, std::size_t y) {
	int res = 0;
	for (std::size_t i : {x - 1, x, x + 1}) {
		for (std::size_t j : {y - 1, y, y + 1}) {
			if (i == x && j == y) {
				continue;
			}
			if (cells[i][j] == cell_state::ALIVE) {
				res += 1;
			}
		}
	}
	return res;
}

// Does one step of Game of Live over inputa argument 
// and returns ne cell matrix
cell_matrix gol_iterate(const cell_matrix& cells) {
	cell_matrix res{cell_state::DEAD};
	for (std::size_t row_i = 1; row_i < cells.size() - 1; ++row_i) {
		const auto& row = cells[row_i];

		for (std::size_t col_i = 1; col_i < row.size() - 1; ++col_i) {
			int l = liveness(cells, row_i, col_i);
			const cell_state& c = row[col_i];
			if (c == cell_state::ALIVE && (l == 2 || l == 3)) {
				res[row_i][col_i] = cell_state::ALIVE;
			} else if (c == cell_state::DEAD && l == 3) {
				res[row_i][col_i] = cell_state::ALIVE;
			} else {
				res[row_i][col_i] = cell_state::DEAD;
			}
		}
	}
	return res;
}

cell_matrix& get_cells(lua_State* L) {
	return *static_cast<cell_matrix*>(
	    lua_touserdata(L, lua_upvalueindex(1)));
}
gol_window& get_gui(lua_State* L) {
	return *static_cast<gol_window*>(
	    lua_touserdata(L, lua_upvalueindex(2)));
}

int lua_cells_get(lua_State* L) {
	cell_matrix& cells = get_cells(L);

	int x = luaL_checknumber(L, 1);
	int y = luaL_checknumber(L, 2);

	lua_pushnumber(L, static_cast<lua_Number>(cells[x][y]));
	return 1;
}

int lua_cells_set(lua_State* L) {
	cell_matrix& cells = get_cells(L);

	int x = luaL_checknumber(L, 1);
	int y = luaL_checknumber(L, 2);
	int val = luaL_checknumber(L, 3);

	cells[x][y] = static_cast<cell_state>(val);
	return 0;
}

int lua_gol_iterate(lua_State* L) {
	cell_matrix& cells = get_cells(L);
	cells = gol_iterate(cells);
	return 0;
}

int lua_draw(lua_State* L) {
	cell_matrix& cells = get_cells(L);

	gol_window& gui = get_gui(L);
	gui.draw_cell_matrix(cells);
	return 0;
}

int lua_sleep(lua_State*) {
	SDL_Event event;
	while (SDL_PollEvent(&event) == 1) {
		if (event.type == SDL_QUIT) {
			std::exit(0);
		}
	}
	SDL_Delay(100);
	return 0;
}

void register_function(lua_State* L, gol_window& gui, cell_matrix& cells,
		       std::string name, lua_CFunction fun) {
	lua_pushlightuserdata(L, &cells);
	lua_pushlightuserdata(L, &gui);
	lua_pushcclosure(L, fun, 2);
	lua_setglobal(L, name.c_str());
}

int main(int argc, const char* argv[]) {
	if (SDL_Init(SDL_INIT_VIDEO) == -1) {
		std::cerr << "Failed to initialize SDL";
		return 1;
	}
	if (argc != 2) {
		std::cerr << "Provide filename for arg";
		return 1;
	}

	gol_window gui{};
	cell_matrix cells{cell_state::DEAD};

	lua_State* L = luaL_newstate();
	luaL_openlibs(L);

	lua_pushnumber(L, gol_window::cells_width);
	lua_setglobal(L, "width");

	lua_pushnumber(L, gol_window::cells_height);
	lua_setglobal(L, "height");

	register_function(L, gui, cells, "get", lua_cells_get);
	register_function(L, gui, cells, "set", lua_cells_set);
	register_function(L, gui, cells, "draw", lua_draw);
	register_function(L, gui, cells, "gol_iterate", lua_gol_iterate);
	register_function(L, gui, cells, "sleep", lua_sleep);

	int status = luaL_loadfile(L, argv[1]);
	if (status != 0) {
		std::cerr << "Failed to load lua file: " << lua_tostring(L, -1);
		return 3;
	}
	status = lua_pcall(L, 0, 0, 0);
	if (status != 0) {
		std::cerr << "Failed to execute lua file: "
			  << lua_tostring(L, -1);
		return 4;
	}
	SDL_Quit();
}
